'''Deal with Swift.'''

import swiftclient.client
from swiftclient.exceptions import ClientException

from ..util import CountFile, retry, logger
from ..exceptions import UploadException, DownloadException


class Swift(object):
    '''Our connection to S3'''
    # The size of the chunk to download / upload
    chunk_size = 1024 * 1024

    def __init__(self, *args, **kwargs):
        # We explicitly disable retries so that we can manage that directly
        kwargs['retries'] = 0
        self.conn = swiftclient.client.Connection(*args, **kwargs)

    def download(self, bucket, key, fobj, retries, headers=None):
        '''Download the contents of bucket/key to fobj'''
        # Make a file that we'll write into
        fobj = CountFile(fobj)

        # Get its original location so we can go back to it if need be
        offset = fobj.tell()

        @retry(retries)
        def func():
            '''The bit that we want to retry'''
            try:
                resp_headers, response = self.conn.get_object(
                    bucket, key, resp_chunk_size=self.chunk_size, headers=headers)

                fobj.seek(offset)
                for chunk in response:
                    fobj.write(chunk)

                length = resp_headers.get('content-length')
                if not length:
                    logger.warn('No content-length provided -- cannot detect truncation.')
                elif fobj.count != int(length):
                    raise DownloadException('Downloaded only %i of %i bytes' % (
                        fobj.count, int(length)))

            except ClientException as exc:
                raise DownloadException('Key not found: ' + str(exc))

        # Invoke the download
        func()

    def upload(self, bucket, key, fobj, retries, headers=None):
        '''Upload the contents of fobj to bucket/key with headers'''
        # Make our headers object
        headers = headers or {}

        @retry(retries)
        def func():
            try:
                self.conn.put_object(
                    bucket, key, fobj, chunk_size=self.chunk_size, headers=headers)
            except ClientException:
                raise UploadException('Failed to upload %s' % key)

        func()

    def _get_container_retry(self, retries, *args, **kwargs):
        '''Wrap Swift's get_container with retries.'''
        @retry(retries)
        def func():
            return self.conn.get_container(*args, **kwargs)
        return func()

    def list(self, bucket, prefix=None, delimiter=None, retries=3,
                   headers=None, chunksize=100):
        '''List the bucket, possibly limiting the search with a prefix.'''
        listing =  self._get_container_retry(retries,
                                           bucket,
                                           prefix=prefix,
                                           delimiter=delimiter,
                                           limit=chunksize)[1]
        while listing:
            for result in listing:
                yield result['name']
            # out of results in current listing, get more starting at the end
            marker = listing[-1]['name']
            listing = self._get_container_retry(retries,
                                              bucket,
                                              prefix=prefix,
                                              delimiter=delimiter,
                                              marker=marker,
                                              limit=chunksize)[1]



'''Batching with gevent'''

import sys
if 'threading' in sys.modules:
    del sys.modules['threading']
from gevent import monkey
monkey.patch_all()
from gevent.pool import Pool


class Proxy(object):
    '''A proxy that will run a function on a new connection in a gevent pool'''
    def __init__(self, pool, func):
        self.pool = pool
        self.func = func
        self._greenlet = None

    def run(self, *args, **kwargs):
        '''Invoke our function with arguments'''
        callback = kwargs.pop('callback', None)
        if callback:
            return callback(self.func(*args, **kwargs))
        return self.func(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self._greenlet = self.pool.spawn(self.run, *args, **kwargs)
        return self._greenlet

    def __getattr__(self, attr):
        return getattr(self._greenlet, attr)


class Batch(object):
    '''For uploading batches of objects in parallel. Implements the same
    interface as Connection, but just does it all in parallel. Can be used as a
    ContextManager'''

    def __init__(self, connection, poolsize):
        # Save a copy of connection, and the pool size
        self.conn = connection
        self.pool = Pool(poolsize)
        self.proxies = []

    def __getattr__(self, attr):
        # Return a proxy object that will perform the same action in a pool
        proxy = Proxy(self.pool, getattr(self.conn, attr))
        self.proxies.append(proxy)
        return proxy

    def __enter__(self):
        return self

    def __exit__(self, typ, val, trace):
        # Wait for all the jobs in the pool to complete
        self.wait()
        if typ:   # pragma: no cover
            raise typ, val, trace

    def wait(self):
        '''Wait until all our jobs are done'''
        self.pool.join()

    def success(self):
        '''Return whether or not everything finished successfully'''
        return not any(not p.successful() for p in self.proxies)

    def results(self):
        '''Get the results of each request, in original order'''
        return list(p.value for p in self.proxies)



'''Various utilities'''

import time

# Logging
import logging
logger = logging.getLogger('s3po')
formatter = logging.Formatter('[%(asctime)s] %(levelname)s in %(module)s:%(funcName)s@%(lineno)s => %(message)s')
handler = logging.StreamHandler()
handler.setLevel(logging.WARNING)
handler.setFormatter(formatter)
logger.addHandler(handler)


class CountFile(object):
    '''A file-like object that counts how much data's written to it'''
    def __init__(self, fobj):
        self._fobj = fobj
        self._start = fobj.tell()

    @property
    def count(self):
        '''How many bytes we've written'''
        return self._fobj.tell() - self._start

    def __getattr__(self, attr):
        return getattr(self._fobj, attr)


class Backoff(object):
    '''Various backoff policies'''
    @staticmethod
    def exponential(factor, base):
        '''Exponential backoff of the form factor * (base ^ x)'''
        def func(num):
            '''Exponential backoff of the form factor * (base ^ x)'''
            return factor * (base ** num)
        return func


def retry(count,
    exceptions=(Exception), sleep=None, policy=Backoff.exponential(30, 2)):
    '''Decorator for retrying a function count times'''
    sleep = sleep or time.sleep

    def _retry(func):
        '''The actual decorator'''
        def new_func(*args, **kwargs):
            '''The decorated function'''
            for attempt in xrange(count + 1):  # pragma: no branch
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    if attempt >= (count - 1):
                        raise
                    interval = policy(attempt)
                    logger.exception(
                        'Sleeping %is after attempt %i' % (interval, attempt))
                    sleep(interval)
        return new_func
    return _retry



#!/usr/bin/env python

# Copyright (c) 2011, 2013 SEOmoz, Inc
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from setuptools import setup

setup(
    name             = 's3po',
    version          = '0.4.6',
    description      = 'An uploading daemon for S3',
    long_description = '''Boto is a wonderful library. This is just a little
        help for dealing with multipart uploads, batch uploading with gevent
        and getting some help when mocking''',
    author           = 'Dan Lecocq',
    author_email     = 'dan@moz.com',
    url              = 'http://github.com/seomoz/s3po',
    packages         = ['s3po', 's3po.backends'],
    license          = 'MIT',
    platforms        = 'Posix; MacOS X',
    install_requires = [
        'boto',
        'coverage',
        'gevent',
        'mock',
        'nose',
        'python_swiftclient'
    ],
    classifiers      = [
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP'
    ]
)



'''Deal with object storage.'''

import contextlib
import os
from cStringIO import StringIO

# Internal imports
from .util import retry, logger
from .backends.s3 import S3
from .backends.swift import Swift
from .backends.memory import Memory


class Connection(object):
    '''Our connection to S3'''

    @classmethod
    def s3(cls, *args, **kwargs):  # pragma: no cover
        '''Create a connection using S3.'''
        return cls(S3(*args, **kwargs))

    @classmethod
    def swift(cls, *args, **kwargs):  # pragma: no cover
        '''Create a connection using Swift.'''
        return cls(Swift(*args, **kwargs))

    @classmethod
    def memory(cls):
        '''Create a connection using the in-memory backend.'''
        return cls(Memory())

    def __init__(self, backend):
        self.backend = backend

    def batch(self, poolsize=20):
        from .batch import Batch
        return Batch(self, poolsize)

    @contextlib.contextmanager
    def mock(self):
        '''Return a context-manager for managing our mocking'''
        original = self.backend
        try:
            self.backend = Memory()
            yield self
        finally:
            self.backend = original

    def upload(self, bucket, key, obj_or_data, headers=None, retries=3):
        '''Upload the provided string or file object to bucket/key'''
        logger.info('Uploading to %s / %s', bucket, key)
        if isinstance(obj_or_data, basestring):
            return self.backend.upload(
                bucket, key, StringIO(obj_or_data), retries, headers)
        else:
            return self.backend.upload(bucket, key, obj_or_data, retries, headers)

    def upload_file(self, bucket, key, path, headers=None, retries=3):
        '''Upload the file at path to bucket/key. This method is important for
        use in batch mode, so that the file object can be used with the right
        context management'''
        with open(os.path.abspath(path)) as fobj:
            return self.upload(bucket, key, fobj, headers, retries)

    def download(self, bucket, key, obj=None, headers=None, retries=3):
        '''Download to either the object or return a string'''
        logger.info('Downloading %s / %s', bucket, key)
        if obj:
            return self.backend.download(bucket, key, obj, retries, headers)
        obj = StringIO()
        self.backend.download(bucket, key, obj, retries, headers)
        return obj.getvalue()

    def download_file(self, bucket, key, path, headers=None, retries=3, mode='w'):
        '''Download the item at bucket/key to a file at path. This method is
        important for us in batch mode so that the file object can be used with
        the right context management'''
        with open(os.path.abspath(path), mode) as fout:
            return self.download(bucket, key, fout, headers, retries)

    def list(self, bucket, prefix=None, delimiter=None, retries=3, headers=None):
        '''List the contents of the bucket, optionally specifying a prefix.'''
        return self.backend.list(bucket, prefix, delimiter, retries, headers)
